__title__ = "MindsDB Confluence handler"
__package_name__ = "mindsdb_confluence_handler"
__version__ = "0.0.1"
__description__ = "MindsDB handler for Confluence"
__author__ = "Balaji Seetharaman"
__github__ = "https://github.com/mindsdb/mindsdb"
__pypi__ = "https://pypi.org/project/mindsdb/"
__license__ = "MIT"
__copyright__ = "Copyright 2023 - mindsdb"
