"""predictor_integration_null

Revision ID: ee63d868fa84
Revises: b6d0a47294ac
Create Date: 2023-02-17 13:48:49.464644

"""
from alembic import op
import sqlalchemy as sa
import mindsdb.interfaces.storage.db # noqa


# revision identifiers, used by Alembic.
revision = 'ee63d868fa84'
down_revision = 'b6d0a47294ac'
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###

    with op.batch_alter_table('predictor', schema=None) as batch_op:
        batch_op.alter_column(
            'integration_id',
            existing_type=sa.INTEGER(),
            nullable=True
        )

    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table('predictor', schema=None) as batch_op:
        batch_op.alter_column(
            'integration_id',
            existing_type=sa.INTEGER(),
            nullable=False
        )

    # ### end Alembic commands ###
