# Default settings for markdown header splitting
DEFAULT_MARKDOWN_HEADERS = [
    ("#", "Header 1"),
    ("##", "Header 2"),
    ("###", "Header 3"),
]

# Limits for web crawling
DEFAULT_CRAWL_DEPTH = 1
DEFAULT_WEB_CRAWL_LIMIT = 100
DEFAULT_WEB_FILTERS = []

DEFAULT_CONTEXT_DOCUMENT_LIMIT = 50
